#!/usr/bin/perl -w
#

# Parse JSON stats from BTC Guild for Cacti
#
# Can be invoked in three seperate modes for each of the stats: User, Miners and Pools
#
# Sample invocation:
#
# User
#   query_btcguild_stats.pl HOST API_KEY user 
#
# Miners
#   query_btcguild_stats.pl HOST API_KEY miners index
#
# Pools
#   query_btcguild_stats.pl HOST API_KEY pool index
#
######################
## Revision History ##
######################
# 
# v0.04 7/5/2011 JinTu (jintuNO@SPAMpraecogito.com)
#	Minor bugfixes for worker indexing after the addition of worker totals. Added more specific memcached namespace. Added simple check for when the API JSON gives us null.
#
# v0.03 7/3/2011 JinTu (jintuNO@SPAMpraecogito.com)
#	Added worker totals
#
# v0.02 6/29/2011 JinTu (jintuNO@SPAMpraecogito.com)
#	Added simple JSON caching via memcached resulting in big performance boost:
#        Before -> SYSTEM STATS: Time:114.1648 Method:spine Processes:2 Threads:4 Hosts:12 HostsPerProcess:6 DataSources:159 RRDsProcessed:105
#        After  -> SYSTEM STATS: Time:19.5116 Method:spine Processes:2 Threads:4 Hosts:12 HostsPerProcess:6 DataSources:159 RRDsProcessed:105
#
# v0.01	6/24/2011 JinTu (jintuNO@SPAMpraecogito.com)
#	First working version. Only supports btcguild stats for now. No JSON query caching support (Slow!). 
# 

use strict;
use JSON -support_by_pp;
use HTTP::Request;
use LWP::UserAgent;

my $memcached_server;
# Uncomment the following three lines and set $memcached_server as appropriate if you are using memcached
#use Cache::Memcached;
#use IO::Socket::INET;
#$memcached_server = "localhost:11211";
#
# The API JSON check interval in seconds. Should be slightly less than your polling interval to ensure fresh data each polling cycle
my $check_interval = 150;
# The maximum amount of time to wait (in seconds) for the lock loop to wait before baling out and attempting to fix things
my $max_loop_time = 60;
#
###### Do not edit below here unless you know what you are doing
#

my $host = $ARGV[0];
my $api_key = $ARGV[1];
my $mode = $ARGV[2];

unless ((defined $host) && (defined $api_key) && (($ARGV[2] eq "user") || ($ARGV[2] eq "workers") || ($ARGV[2] eq "pool") || ($ARGV[2] eq "total_all_pools"))) {
  # Missing input params
  print "usage:\n\n./query_btcguild_stats.pl HOST API_KEY user\n\n" .
        "./query_btcguild_stats.pl HOST API_KEY workers index\n" . 
        "./query_btcguild_stats.pl HOST API_KEY workers num_indexes\n" . 
        "./query_btcguild_stats.pl HOST API_KEY workers query {worker_name,hash_rate,round_shares,round_stales,reset_shares,reset_stales,total_shares,total_stales,last_share,blocks_found}\n" . 
        "./query_btcguild_stats.pl HOST API_KEY workers get {worker_name,hash_rate,round_shares,round_stales,reset_shares,reset_stales,total_shares,total_stales,last_share,blocks_found} WORKER\n\n" . 
        "./query_btcguild_stats.pl HOST API_KEY pool index\n" .
        "./query_btcguild_stats.pl HOST API_KEY pool num_indexes\n" .
        "./query_btcguild_stats.pl HOST API_KEY pool query {pool_name,speed}\n" .
        "./query_btcguild_stats.pl HOST API_KEY pool get {pool_name,speed} POOL\n\n" .
        "./query_btcguild_stats.pl HOST API_KEY total_all_pools\n";
  exit 0;
}

my $json_url = "https://$host/api.php?api_key=$api_key";
my $content;

eval{
    # download the json page:
    if (defined $memcached_server) {
       # Memcached is enabled
       #print "Using memcached\n";
       my $msock = IO::Socket::INET->new( PeerAddr => $memcached_server,
                                          Timeout  => 3);
       unless ($msock) {
          print "Unable to connected to memcached server at $memcached_server, aborting!\n";
          exit 0;
       }
       my $memd = new Cache::Memcached {
               'servers' => [ $memcached_server ],
               'debug' => 0,
               'compress_threshold' => 10_000,
               'namespace' => "Cacti::BTCGuild/$host/$api_key/",
       };
       # Check the timestamp
       my $timestamp = $memd->get("timestamp");
       my $api_json = $memd->get("api_json");
       if ((defined $timestamp) && (defined $api_json)) {
          # Check if we are beyond the check interval based on the last timestamp
          if (time() >= ($timestamp + $check_interval)) {
             # We have exceeded the check interval and might need to fetch an update if no one else is.
             unless (defined $memd->get("lock")) {
                $memd->set("lock", 1);
                $content = fetch_json($json_url);
                $memd->set("api_json", $content);
                $memd->set("timestamp", time());
                $memd->delete("lock");
             }
             else {
                #print "Locked\n";
                # Loop for a while waiting for the lock to clear
                my $time_elapsed = 0;
                while (defined $memd->get("lock")) {
                   #print "locked, looping. time_elapsed = $time_elapsed\n";
                   # sleep for a random amount of time from 100-600ms
                   my $wait_time = 0.1 + (int(rand(500)) * 0.001);
                   select(undef, undef, undef, $wait_time);
                   $time_elapsed += $wait_time;
                   if ($time_elapsed >= $max_loop_time) {
                      # Loop timer elapsed. Something has gone wrong
                      $content = fetch_json($json_url);
                      $memd->set("api_json", $content);
                      $memd->set("timestamp", time());
                      $memd->delete("lock");
                   }
                }
                # Assume that now that the lock is cleared, the data is current
                #print "Lock loop has exited\n";
                $content = $api_json;
             }
          }
          else {
             #print "Data in cache is current:" . time() . " > " . ($timestamp + $check_interval) . "\n";
             # We don't need to check again, just use the cached value
             $content = $api_json;
          }
       }
       else {
          # timestamp and/or json aren't defined (e.g. first run). We need to fetch the JSON and set the timestamp
          $memd->set("lock", 1);
          $content = fetch_json($json_url);
          $memd->set("api_json", $content);
          $memd->set("timestamp", time());
          $memd->delete("lock");
       }
    }
    else {
       # not using memcached, so fetch every time
       $content = fetch_json($json_url);
    }
    # Simple validation prior to parsing
    if ($content eq "") {
      print "$host API returned null, not valid JSON!\n";
      exit 0;
    }
    my $json = new JSON;
 
    # these are some nice json options to relax restrictions a bit:
    my $json_text = $json->allow_nonref->utf8->relaxed->escape_slash->loose->allow_singlequote->allow_barekey->decode($content);
     
    # Cacti needs whole numbers, so this needs to be divided by 100000000 to get the original value and retain full 8 digit precision
    if ($mode eq "user") {
       print "confirmed_rewards:" . $json_text->{user}->{confirmed_rewards} * 100000000 . " " .
        "unconfirmed_rewards:" . $json_text->{user}->{unconfirmed_rewards} * 100000000 . " " .
        "estimated_rewards:" . $json_text->{user}->{estimated_rewards} * 100000000 . " " .
        "payouts:" . $json_text->{user}->{payouts} * 100000000;
    }
    elsif($mode eq "workers") {
      # Enumerate the worker names so we don't have indexing issues and get sums for total
      my %workerid;
      my %workertotal;
      foreach my $worker(%{$json_text->{workers}}) {
        if (defined $json_text->{workers}->{$worker}->{worker_name}) {
          $workerid{$json_text->{workers}->{$worker}->{worker_name}} = $worker;
          foreach my $wparam (keys %{$json_text->{workers}->{$worker}}) {
            unless ($wparam eq "worker_name") {
              if (defined $workertotal{$wparam}) {
                if ($wparam eq "last_share") {
                  $workertotal{$wparam} += convert_time($json_text->{workers}->{$worker}->{$wparam});
                }
                else {
                  $workertotal{$wparam} += $json_text->{workers}->{$worker}->{$wparam};
                }
              } 
              else {
                if ($wparam eq "last_share") {
                  $workertotal{$wparam} = convert_time($json_text->{workers}->{$worker}->{$wparam});
                }
                else {
                  $workertotal{$wparam} = $json_text->{workers}->{$worker}->{$wparam};
                }
              }
              #print "$wparam for " . $json_text->{workers}->{$worker}->{worker_name} . " set to " . $json_text->{workers}->{$worker}->{$wparam} . "\n";
            }
          }
        }
      }     
      if (($ARGV[3] eq "get") && (defined $ARGV[4]) && (defined $ARGV[5])) {
        if ($ARGV[5] eq "total") {
          if ($ARGV[4] eq "hash_rate") {
            # need to make this a whole number for Cacti
            printf ("%d", $workertotal{$ARGV[4]});
          }
          else {
            print $workertotal{$ARGV[4]};
          }
        }
        elsif (defined $json_text->{workers}->{$workerid{$ARGV[5]}}->{$ARGV[4]}) {
          if ($ARGV[4] eq "hash_rate") {
            # need to make this a whole number for Cacti
            printf ("%d", $json_text->{workers}->{$workerid{$ARGV[5]}}->{$ARGV[4]});  
          }
          elsif ($ARGV[4] eq "last_share") {
            print convert_time($json_text->{workers}->{$workerid{$ARGV[5]}}->{$ARGV[4]});
          }
          else {
            print $json_text->{workers}->{$workerid{$ARGV[5]}}->{$ARGV[4]};
          }
        }
      }
      elsif (($ARGV[3] eq "query") && (defined $ARGV[4])) {
        foreach my $worker_name (keys %workerid) {
          if (($ARGV[4] eq "worker_name") && (defined $json_text->{workers}->{$workerid{"$worker_name"}}->{worker_name})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{worker_name}\n";
          }
          elsif (($ARGV[4] eq "hash_rate") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{hash_rate})) {
      	    print "$worker_name:" . sprintf("%d",$json_text->{workers}->{$workerid{$worker_name}}->{hash_rate}) ."\n";
          }
          elsif (($ARGV[4] eq "round_shares") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{round_shares})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{round_shares}\n";
          }
          elsif (($ARGV[4] eq "round_stales") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{round_stales})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{round_stales}\n";
          }
          elsif (($ARGV[4] eq "reset_shares") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{reset_shares})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{reset_shares}\n";
          }
          elsif (($ARGV[4] eq "reset_stales") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{reset_stales})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{reset_stales}\n";
          }
          elsif (($ARGV[4] eq "total_shares") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{total_shares})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{total_shares}\n";
          }
          elsif (($ARGV[4] eq "total_stales") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{total_stales})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{total_stales}\n";
          }
          elsif (($ARGV[4] eq "last_share") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{last_share})) {
      	    print "$worker_name:" . convert_time($json_text->{workers}->{$workerid{$worker_name}}->{last_share}) ."\n";
          }
          elsif (($ARGV[4] eq "blocks_found") && (defined $json_text->{workers}->{$workerid{$worker_name}}->{blocks_found})) {
      	    print "$worker_name:$json_text->{workers}->{$workerid{$worker_name}}->{blocks_found}\n";
          }
        }
        if ($ARGV[4] eq "worker_name") {
          print "total:total\n";
        }
        elsif ($ARGV[4] eq "hash_rate") {
           print "total:" . sprintf("%d",$workertotal{$ARGV[4]}) . "\n";
        }
        else {
          print "total:" . $workertotal{$ARGV[4]} . "\n";
        }
      }
      elsif ($ARGV[3] eq "index") {
        foreach my $worker_name (keys %workerid) {
          print "$worker_name\n";
        }
        print "total\n";
      }
      elsif ($ARGV[3] eq "num_indexes") {
        print scalar(keys %workerid) + 1;
      }
    }
    elsif ($mode eq "pool") {
      if (($ARGV[3] eq "get") && (defined $json_text->{pool}->{$ARGV[5] . "_speed"})) {
         if ($ARGV[4] eq "speed") {
            printf ("%d", $json_text->{pool}->{$ARGV[5] . "_speed"});
         }
         elsif ($ARGV[4] eq "pool_name") {
            print $ARGV[5];
         }
      }
      else {
        my $num_pools = 0;
        foreach my $pool (%{$json_text->{pool}}) {
           if (my ($pname) = $pool =~ m/(.*)_speed/) {
              if ($ARGV[3] eq "index") {
                 print "$pname\n";
              }
              elsif ($ARGV[3] eq "query") {
                 if ($ARGV[4] eq "speed") {
                    print "$pname:" . sprintf ("%d", $json_text->{pool}->{$pname . "_speed"}) . "\n"; 
                 }
                 elsif ($ARGV[4] eq "pool_name") {
                    print "$pname:$pname\n";
                 }
              }
              $num_pools++;
           }
        }
        if ($ARGV[3] eq "num_indexes") {
           print $num_pools;
        }
      }
    }
    elsif ($mode eq "total_all_pools") {
      print "hash_rate:" . sprintf ("%d", $json_text->{pool}->{hash_rate}) . " " .
         "active_workers:$json_text->{pool}->{active_workers} " .
         "round_shares:$json_text->{pool}->{round_shares} " .
         "round_time:" . convert_time($json_text->{pool}->{round_time});
    }
  };
  # catch crashes:
  if($@){
    print "Crash detected: $@\n";
  }

# Convert HH:MM:SS into seconds
sub convert_time {
   my ($time_in) = @_;
   my ($hours,$minutes,$seconds) = $time_in =~ m/(\d*):(\d*):(\d*)/;
   return $seconds + ($minutes * 60) + ($hours * 60 * 60); 
}

# Get the API JSON
sub fetch_json { 
   my ($url) = @_;
   my $request = HTTP::Request->new(GET => $url);
   my $ua = LWP::UserAgent->new;
   my $response = $ua->request($request);
   # Should do some error checking
   return $response->content();
}
